(**
 * arguments management
 *)
module Args =
  struct

    let get = List.tl (Array.to_list Sys.argv)

    let is_primary arg = List.mem ("-" ^ arg) get

    let get_primary_val main =
      let rec aux l ok =
	match l with
	  [] -> []
	| a :: l ->
	    if a = ("-" ^ main) then aux l true
	    else if String.get a 0 = '-' then aux l false
	    else
	      let subl = aux l ok in
	      if ok then a :: subl else subl in
      aux get false

    let is_secondary' main arg = List.mem arg (get_primary_val main)

    (**
     * displays the program command usage and exits.
     *)
    let usage _ =
      (print_string ("Usage: dbgen <options> -i <in-file> -o <out-file>\n");
       print_string ("Options are:\n");
       print_string ("  -g -gen <list>\tAdditionnal generation according to <list>:\n");
       print_string ("    named-nat\t\tnamed syntax with nat variables\n");
       print_string ("    named-string\tnamed syntax with string variables\n");
       print_string ("  -d -debug\tPrint internal execution trace and state\n");
       print_string ("  -v -version\tPrint version and exit\n");
       exit Sys.sigabrt)

    let authorized_args =
      "-v" :: "-version" :: "-d" :: "-debug" :: "-g" :: "-gen" ::
      "-i" :: "-o" :: []

    let check_arg arg =
      if not (List.mem arg authorized_args) then
	(print_string ("Unrecognized argument: " ^ arg); usage ())
      else ()
      

	  (** Accessors to concrete arguments *)
    let version_invoked _ =
      (is_primary "v") || (is_primary "version")

    let get_input_file_name _ =
      let args = get_primary_val "i" in
      if List.length args <> 1 then usage () else List.hd args

    let get_output_file_name _ =
      let args = get_primary_val "o" in
      if List.length args <> 1 then usage () else List.hd args

    let is_aux arg = is_secondary' "gen" arg || is_secondary' "g" arg

  end
;;
