open Xlist
open Xstring
open Args

open Interfaces
open Definitions_abstract
open Definitions_lift_subst

open SizedList

class fv_permut_fun _gmod =
  object (self)
    inherit lift_fun _gmod
	
	(** the definition title *)
    method title : string = "Free variable index permutations"
	(** function parameters list (name * type_name) *)
    method fun_params : (string * string) list =
      ("_k", "nat") ::
      ("_l", "list nat") ::
      ("_m", "nat") :: []
			
			(** radix of the function name *)
    method fun_radix : string = "fv_permut"
	
    method index_tactic_name : string option = Some "fv_permut_index_tac"
						      
						      (** generation of parameters *)
    method private parameter_rhs_function_gen (p : parameter) : string =
      let shift =
	let raw_shifts = List.map self#shift_rhs_function_gen p#get_shifts in
	let non_empty_shifts = List.filter (fun s -> s <> "") raw_shifts in
	String.concat " + " non_empty_shifts in
      let subst =
	("#param#", p#get_name) ::
	("#recfun#", self#fun_name_type_name p#get_type#get_name) ::
	("#shift#", if shift = "" then "_m" else "(_m + " ^ shift ^ ")") ::
	[]
      and text =
	if self#is_self_rec p || self#is_back_rec p || self#is_throughs_rec p
	then
	  "(#recfun# _k _l #shift# #param#)"
	else
	  "#param#" in
      Xstring.replace subst text
	
	(** generation of constructors rhs : index case *)
    method private index_constructor_rhs_function_gen (c : constructor) : string list =
      let id_params =
	String.concat " "
	  (List.map (fun p -> self#parameter_rhs_function_gen p) c#get_params) in
      let mod_params_lift =
	String.concat " "
	  (List.map (fun p ->
	    if p#is_index then "(" ^ p#get_name ^ " + _k)"
	    else self#parameter_rhs_function_gen p) c#get_params) in
      let mod_params_nth =
	String.concat " "
	  (List.map (fun p ->
	    if p#is_index
	    then "(_m + List.nth (" ^ p#get_name ^ " - _m) _l 0)"
	    else self#parameter_rhs_function_gen p) c#get_params) in
      let subst =
	("#target_type#", self#get_target#get_name) ::
	("#constr#", c#get_name) ::
	("#index#", c#get_index_name) ::
	("#id_params#", id_params) ::
	("#mod_params_lift#", mod_params_lift) ::
	("#mod_params_nth#", mod_params_nth) ::
	[]
      and text =
	if self#get_target = self#get_source
	then
	  "if le_lt_dec _m #index#" ::
	  "then" ::
	  "  if le_lt_dec (List.length _l) (#index# - _m)" ::
	  "  then #constr# #mod_params_lift#" ::
	  "  else #constr# #mod_params_nth#" ::
	  "else #constr# #id_params#" ::
	  []
	else "#constr# #index#" :: [] in
      Xstring.replaces subst [] text

	(** generation of index lemmas and tactics *)
    method index_constructor_lemmas_and_tactics (c : constructor) : (string list * string list) =
      let id_params =
	String.concat " "
	  (List.map (fun p -> self#parameter_rhs_function_gen p) c#get_params) in
      let mod_params_lift =
	String.concat " "
	  (List.map (fun p ->
	    if p#is_index then "(" ^ p#get_name ^ " + _k)"
	    else self#parameter_rhs_function_gen p) c#get_params) in
      let mod_params_nth =
	String.concat " "
	  (List.map (fun p ->
	    if p#is_index
	    then "(_m + List.nth (" ^ p#get_name ^ " - _m) _l 0)"
	    else self#parameter_rhs_function_gen p) c#get_params) in
      let params = String.concat " " (List.map (fun p -> p#get_name) c#get_params) in
      let params_why = String.concat " " (List.map (fun p -> "?" ^ p#get_name) c#get_params) in
      let subst =
	("#fv_permut#", self#fun_name) ::
	("#target_type#", self#get_target#get_name) ::
	("#constr#", c#get_name) ::
	("#index#", c#get_index_name) ::
	("#id_params#", id_params) ::
	("#mod_params_lift#", mod_params_lift) ::
	("#mod_params_nth#", mod_params_nth) ::
	("#params#", params) ::
	("#why_params#", params_why) ::
	[]
      and text =
	"  Lemma #fv_permut#_#constr#_dec :" ::
	"    forall (_k : nat) (_l : list nat) (_m : nat) #params#," ::
	"    { prod (prod (_m <= #index#) ((#index# - _m) < List.length _l)) (#fv_permut# _k _l _m (#constr# #params#) = #constr# #mod_params_nth#) } +" ::
	"    { prod (prod (_m <= #index#) (List.length _l <= (#index# - _m))) (#fv_permut# _k _l _m (#constr# #params#) = #constr# #mod_params_lift#) } +" ::
	"    { prod (#index# < _m) (#fv_permut# _k _l _m (#constr# #params#) = #constr# #id_params#) }." ::
	"  Proof." ::
	"    intros; simpl; destruct le_lt_dec; try destruct le_lt_dec; subst; simpl; auto." ::
	"  Qed." ::
	"" ::
	"  Ltac _#fv_permut#_#constr#_dec_tac _k _l _m #params# :=" ::
	"    let Hcomp := fresh \"Hcomp\" in" ::
	"       let Hcomp' := fresh \"Hcomp'\" in" ::
	"          let Hfun := fresh \"Hfun\" in" ::
	"            destruct (#fv_permut#_#constr#_dec _k _l _m #params#) as" ::
	"              [ [ [ [ Hcomp Hcomp' ] Hfun ] | [ [ Hcomp Hcomp' ] Hfun ] ] | [ Hcomp Hfun ] ];" ::
	"              rewrite ?Hfun in *; clear Hfun." ::
	"" ::
	[]
      and text2 =
	"      | [ |- context [ #fv_permut# ?_k ?_l ?_m (#constr# #why_params#) ] ] =>" ::
	"        _#fv_permut#_#constr#_dec_tac _k _l _m #params#" ::
	"" ::
	"      | [ H : context [ #fv_permut# ?_k ?_l ?_m (#constr# #why_params#) ] |- _ ] =>" ::
	"        _#fv_permut#_#constr#_dec_tac _k _l _m #params#" ::
	"" ::
	[] in
      (Xstring.replaces subst [] text, Xstring.replaces subst [] text2)
	
	(** generation of index equation lemmas *)
    method index_constructor_equation_lemmas (c : constructor) : (string list * string list) =
      let id_params =
	String.concat " "
	  (List.map (fun p -> self#parameter_rhs_function_gen p) c#get_params) in
      let mod_params_lift =
	String.concat " "
	  (List.map (fun p ->
	    if p#is_index then "(" ^ p#get_name ^ " + _k)"
	    else self#parameter_rhs_function_gen p) c#get_params) in
      let mod_params_nth =
	String.concat " "
	  (List.map (fun p ->
	    if p#is_index
	    then "(_m + List.nth (" ^ p#get_name ^ " - _m) _l 0)"
	    else self#parameter_rhs_function_gen p) c#get_params) in
      let params = String.concat " " (List.map (fun p -> p#get_name) c#get_params) in
      let params_why = String.concat " " (List.map (fun p -> "?" ^ p#get_name) c#get_params) in
      let subst =
	("#fv_permut#", self#fun_name) ::
	("#target_type#", self#get_target#get_name) ::
	("#constr#", c#get_name) ::
	("#index#", c#get_index_name) ::
	("#id_params#", id_params) ::
	("#mod_params_lift#", mod_params_lift) ::
	("#mod_params_nth#", mod_params_nth) ::
	("#params#", params) ::
	("#why_params#", params_why) ::
	[]
      and text =
	"  Lemma #fv_permut#_#constr#_le_le :" ::
	"    forall (_k : nat) (_l : list nat) (_m : nat) #params#," ::
	"      _m <= #index# -> List.length _l <= (#index# - _m) -> #fv_permut# _k _l _m (#constr# #params#) = #constr# #mod_params_lift#." ::
	"  Proof." ::
	"    intros; simpl; destruct le_lt_dec; try destruct le_lt_dec; f_equal; omega." ::
	"  Qed." ::
	"" ::
	"  Lemma #fv_permut#_#constr#_le_lt :" ::
	"    forall (_k : nat) (_l : list nat) (_m : nat) #params#," ::
	"      _m <= #index# -> (#index# - _m) < List.length _l -> #fv_permut# _k _l _m (#constr# #params#) = #constr# #mod_params_nth#." ::
	"  Proof." ::
	"    intros; simpl; destruct le_lt_dec; try destruct le_lt_dec; f_equal; omega." ::
	"  Qed." ::
	"" ::
	"  Lemma #fv_permut#_#constr#_lt :" ::
	"    forall (_k : nat) (_l : list nat) (_m : nat) #params#," ::
	"      #index# < _m -> #fv_permut# _k _l _m (#constr# #params#) = #constr# #id_params#." ::
	"  Proof." ::
	"    intros; simpl; destruct le_lt_dec; try destruct le_lt_dec; f_equal; omega." ::
	"  Qed." ::
	"" ::
	[]
      and text2 =
	"      | [ |- context [ #fv_permut# ?_k ?_l ?_m (#constr# #why_params#) ] ] =>" ::
	"          (cutrewrite ->" ::
	"            (#fv_permut# _k _l _m (#constr# #params#) = #constr# #mod_params_lift#);" ::
	"            [ | apply (#fv_permut#_#constr#_le_le _k _l _m #params#); cfail ]) ||" ::
	"          (cutrewrite ->" ::
	"            (#fv_permut# _k _l _m (#constr# #params#) = #constr# #mod_params_nth#);" ::
	"            [ | apply (#fv_permut#_#constr#_le_lt _k _l _m #params#); cfail ]) ||" ::
	"          (cutrewrite ->" ::
	"            (#fv_permut# _k _l _m (#constr# #params#) = #constr# #id_params#);" ::
	"            [ | apply (#fv_permut#_#constr#_lt _k _l _m #params#); cfail ]) ||" ::
	"          #index_dec_tac#" ::
	"" ::
	"      | [ H : context [ #fv_permut# ?_k ?_l ?_m (#constr# #why_params#) ] |- _ ] =>" ::
	"          (cutrewrite ->" ::
	"            (#fv_permut# _k _l _m (#constr# #params#) = #constr# #mod_params_lift#) in H;" ::
	"            [ | apply (#fv_permut#_#constr#_le_le _k _l _m #params#); cfail ]) ||" ::
	"          (cutrewrite ->" ::
	"            (#fv_permut# _k _l _m (#constr# #params#) = #constr# #mod_params_nth#) in H;" ::
	"            [ | apply (#fv_permut#_#constr#_le_lt _k _l _m #params#); cfail ]) ||" ::
	"          (cutrewrite ->" ::
	"            (#fv_permut# _k _l _m (#constr# #params#) = #constr# #id_params#) in H;" ::
	"            [ | apply (#fv_permut#_#constr#_lt _k _l _m #params#); cfail ]) ||" ::
	"          #index_dec_tac#" ::
	"" ::
	[] in
      (Xstring.replaces subst [] text, Xstring.replaces subst [] text2)
  end

class fv_permut_fun_with_source_and_target _gmod _source _target =
  object (self)
    inherit fv_permut_fun _gmod as super
	
    initializer
      begin
	super#set_source _source;
	super#set_target _target
      end
  end

class named_terms _gmod =
  object (self)
    inherit abstract_terms_definition _gmod as super
	
(** check the type of named variables to use *)
    method private is_string =
      if Args.is_aux "named-string" then true
      else if Args.is_aux "named-nat" then false
      else failwith "Definitions_named.named_terms.is_string: unrecognized named option"

	  (** the definition title *)
    method title : string = "Named syntax"
	(** the category of generated index case *)
    method index_type : string =
      if self#is_string then "string" else "nat"
	  (** extra definition lines to be inserted before the grammar definitions *)
    method coq_predefs : string list =
      let dec =
	if self#is_string then "string_dec" else "eq_nat_dec" in
      let typ =
	if self#is_string then "string" else "nat" in
      let invok =
	if self#is_string then "  Require Import String." :: [""] else [""] in
      SizedList.sizedlist @
      "" ::
      invok @
      ("  Definition _name_list := SizedList.sized_list " ^ typ ^ ".") ::
      ("  Definition _xcons := @SizedList.scons " ^ typ ^ ".") ::
      ("  Definition _xnil := @SizedList.snil " ^ typ ^ ".") ::
      ("  Definition _index_of := SizedList.index_of " ^ dec ^ ".") ::
      ("  Definition _concat := @SizedList.app " ^ typ ^ ".") ::
      ("  Definition _build_permut := SizedList.build_permut " ^ dec ^ ".") ::
      "" ::
      []
	
	(** extra definition lines to be inserted after the constructor cases *)
    method aux_defs : string list =
      (* we get the indexed grammar list *)
      let indexed = gmod#get_indexed_grams_through self#get_target in
      
      (* for each indexed grammar, we build the parameter *)
      let param_gen (g : grammar) =
	"(_xn_" ^ g#get_name ^ " : nat) " ^
	"(_xl_" ^ g#get_name ^ " : _name_list _xn_" ^ g#get_name ^ ") " in
      
      (* applying param_gen to every indexed grammars *)
      let params = List.map param_gen indexed in
      
      (* concatenation of params *)
      let all_params = String.concat "" params in
      
      (* output *)
      ("  | _db_#raw_type# " ^ all_params ^ "(_arg : #raw_type#)") ::
      []
	
	(** category names decoration (to avoid conflict) *)
    method decorate_type_name : string -> string =
      fun x -> if gmod#is_gmodule_gram x then "_" ^ x else x
	
	(** constructors names decoration (to avoid conflict) *)
    method decorate_constr_name : string -> string =
      fun x -> "_" ^ x
		       
		       (** generation of shifts *)
    method private shift_terms_gen (gram, level) : string =
      (* we get the indexed grammar list for this target *)
      let indexed = gmod#get_indexed_grams_through self#get_target in
      
      (* we look for the number of type_name among indexed *)
      let nb = Xlist.index_of gram indexed in
      
      (* conversion to string *)
      let nbs = if nb = 0 then "" else string_of_int nb in
      
      (* output *)
      "(xl" ^ nbs ^ " : _name_list " ^ string_of_shift level ^ ")"
						 
						 (** generation of parameters *)
    method parameter_terms_gen (p : parameter) : string =
      super#parameter_terms_gen p
	
  end

class named_to_db_fun _gmod =
  object (self)
    inherit abstract_fun_definition _gmod as super

	(** the definition title *)
    method title : string = "Named to De Bruijn translation function"
	
	(** extra definition lines to be inserted before the grammar definitions *)
    method coq_predefs : string list =
      []

	(** the named syntax generator *)
    method named_terms_gen = new named_terms _gmod
	
	(** extra definition lines to be inserted after the constructor cases *)
    method aux_defs : string list =
      (* we need the fv_permut function for this definition : this generates
	 the function name to use for a given source grammar *)
      let make_fv_permut (source : grammar) (rec_arg : string) : string =
	let gname = source#get_name in
	
	let fv_permut =
	  new fv_permut_fun_with_source_and_target
	    gmod source self#get_target in
	
	fv_permut#fun_name ^ " (_n_" ^ gname ^ " - _xn_" ^ gname ^ ") "
	^ "(_build_permut _n_" ^ gname ^ " _l_" ^ gname
	^ " _xn_" ^ gname ^ " _xl_" ^ gname ^ ")" ^ rec_arg in
      
      (* we get the indexed grammar list for this target *)
      let indexed = gmod#get_indexed_grams_through self#get_target in
      
      (* folds all indexed grammars to be lifted *)
      let rec lift_all_indexed l =
	match l with
	  [] -> "_arg"
	| gram :: l ->
	    "(" ^ make_fv_permut gram (" 0 " ^ lift_all_indexed l) ^ ")" in
      
      (* apply the previous folding to indexed *)
      let rhs = lift_all_indexed indexed in

      (* for each indexed grammar, we build the parameter *)
      let param_gen (g : grammar) =
	"_xn_" ^ g#get_name ^ " _xl_" ^ g#get_name ^ " " in
      
      (* applying param_gen to every indexed grammars *)
      let params = List.map param_gen indexed in
      
      (* concatenation of params *)
      let all_params = String.concat "" params in
      
      (* output *)
      ("      | _db_#raw_target_type# " ^ all_params ^ "_arg =>") ::
      ("        " ^ rhs) ::
      []
	
	(** category names decoration (to avoid conflict) *)
    method decorate_type_name (name : string) : string =
      if gmod#is_gmodule_gram name then "_" ^ name else name
	      
	      (** constructors names decoration (to avoid conflict) *)
    method decorate_constr_name (name : string) : string =
      "_" ^ name
	      
	      (** function parameters list (name * type_name) *)
    method fun_params : (string * string) list =
      (* we get the indexed grammar list for this target *)
      let indexed = gmod#get_indexed_grams_through self#get_target in

      (* for each indexed grams throughs, builds the function parameters *)
      List.concat
	(List.map
	   (fun g ->
	     ("_n_" ^ g#get_name, "nat") ::
	     ("_l_" ^ g#get_name, "_name_list _n_" ^ g#get_name) ::
	     [])
	   indexed)
	
	(** return type of the function *)
    method return_type : string =
      self#get_target#get_name
	
	(** radix of the function name *)
    method fun_radix : string = "named_to_db"
			       
			       (** the source grammars for which generetion is needed.
				* In this case : every indexed grammar. *)
    method get_source_grams : grammar option list =
      None ::
      []
	
	(** the target (processed) grammars for which generetion is needed.
	 * By default, every grammar with the source grammar in its throughs. *)
    method get_target_grams (grams : grammar list) : grammar list =
      grams
	
	(** Unused virtual function *)
    method shift_rhs_function_gen (type_name, level) : string = ""
	
    method parameter_rhs_function_gen (p : parameter) : string =
      failwith "Definitions_others.named_to_db_fun#parameter_function_gen: invalid call"
    method private parameter_rhs_function_gen_bis (p : parameter) (n : int) : string * int =
      (* Generation function for a shift (with given counter for xlN) *)
      let shift_rhs_function_gen (type_name, level) nb : grammar * (string * string) =
	let nbs = if nb = 0 then "" else string_of_int nb in
	(type_name, (string_of_shift level, "xl" ^ nbs)) in
      
      (* we get the result of mapping shift_rhs_function_gen to all shifts *)
      let shifts =
	Xlist.map_count_start shift_rhs_function_gen n p#get_shifts in

      (* we get the indexed grammar list for this parameter *)
      let indexed = gmod#get_indexed_grams_through p#get_type in
      
      (* function that builds the inductive parameter for a given indexed
	 grammar given the shifts for this parameter *)
      let make_args (gram : grammar) : string =
	(* builds the length additional argument " + ?" *)
	let arg_n =
	  try
	    let (_, (level, _)) = List.find (fun (g, _) -> g = gram) shifts in
	    level					
	  with
	    (* if there is no shift for this grammar *)
	    Not_found -> "" in
	
	(* builds the list argument with potential concatenation *)
	let arg_l =
	  try
	    let (_, (level, id)) = List.find (fun (g, _) -> g = gram) shifts in
	    "(_concat " ^ level ^ " " ^ id ^ " " ^ "_n_" ^ gram#get_name ^ " _l_" ^ gram#get_name ^ ")"
	  with
	    (* if there is no shift for this grammar *)
	    Not_found -> "_l_" ^ gram#get_name in

	let new_n =
	  if arg_n = "" then "_n_" ^ gram#get_name
	  else "(" ^ arg_n ^ " + _n_" ^ gram#get_name ^ ")" in
	new_n ^ " " ^ arg_l in

      let ordered_shifts =
	List.map make_args indexed in

      let list_param = String.concat " " ordered_shifts in
      let next_cpt = List.length shifts in
      let subst =
	("#param#", p#get_name) ::
	("#recfun#", self#fun_name_type_name p#get_type#get_name) ::
	("#list_param#", list_param) ::
	[]
      and text =
	if self#is_module_rec p
	then
	  "(#recfun# #list_param# #param#)"
	else
	  "#param#" in
      (Xstring.replace subst text, next_cpt)

	(** generation of constructors rhs : non index case *)
    method non_index_constructor_rhs_function_gen (c : constructor) : string list =
      [ String.concat " "
	  (c#get_name :: (Xlist.map_count_var (fun p -> self#parameter_rhs_function_gen_bis p) c#get_params)) ]
	
	(** generation of constructors rhs : index case *)
    method index_constructor_rhs_function_gen (c : constructor) : string list =
      let mod_params =
	String.concat " "
	  (Xlist.map_count_var (fun p -> fun n ->
	    if p#is_index
	    then "(_index_of " ^ p#get_name ^ " _n_#raw_target_type# _l_#raw_target_type#)", n + 1
	    else self#parameter_rhs_function_gen_bis p n) c#get_params) in
      let subst =
	("#constr#", c#get_name) ::
	("#index#", c#get_index_name) ::
	[]
      and text =
	("#constr# " ^ mod_params) ::
	[] in
      Xstring.replaces subst [] text
	
	(** generation of parameter lhs *)
    method typed_parameter_lhs_function_gen (p : parameter) : (string * string) =
      failwith "Definitions_others.named_to_db_fun#parameter_lhs_function_gen: invalid call"
    method typed_parameters_lhs_function_gen (c : constructor) : (string * string) list =
      let parameter_lhs_function_gen (p : parameter) (nb : int) : (string * string) list * int =
	(* index case *)
	if p#is_index then ([(p#get_name, self#named_terms_gen#index_type)], 1) else
	let shift_lhs_function_gen (type_name, level) n : string * string =
	  let nbs = if n = 0 then "" else string_of_int n in
	  ("xl" ^ nbs, "_name_list " ^ string_of_shift level) in
	let raw_shifts = Xlist.map_count_start shift_lhs_function_gen nb p#get_shifts in
	(raw_shifts @ [(p#get_name, self#decorate_type_name p#get_type#get_name)], List.length raw_shifts) in
      List.concat (Xlist.map_count_var parameter_lhs_function_gen c#get_params)

	(** generation of constructors *)
    method constructor_function_equation_gen (c : constructor) : string list =
      Xstring.replaces [("#raw_target_type#", super#get_target#get_name)] [] (super#constructor_function_equation_gen c)

  end
