%{
open Lexing
open Error

let current_start_pos () =
  ((symbol_start_pos ()).pos_lnum,
   ((symbol_start_pos ()).pos_cnum - (symbol_start_pos ()).pos_bol))

let current_end_pos () =
  ((symbol_end_pos ()).pos_lnum,
   ((symbol_end_pos ()).pos_cnum - (symbol_end_pos ()).pos_bol))

let current_pos () : Error.region =
  (current_start_pos (), current_end_pos ())

let parse_error s = Error.fail (current_pos ()) s

open Interfaces

open Parameter
open Constructor
open Grammar
open Node
open Gmodule

let gmod = new gmodule_impl "" [] [] []
%}

%token <string> ID
%token <int> INTEGER
%token <string list> VERBATIM_HEADER
%token <string list> VERBATIM_FOOTER

%token EOL

%token MODULE BEGIN END VERBATIM
%token INDUCTIVE TYPE WITH INDEX BIND IN

%token DOT LPARSTAR STARRPAR COLON BAR LPAR RPAR
%token COMMA LSQUARE RSQUARE ASSIGN
%token PLUS MINUS MULT

%start gmod

%type <Interfaces.gmodule> gmod
%type <Interfaces.node list> nodes
%type <Interfaces.node> node
%type <Interfaces.grammar list> grams
%type <Interfaces.constructor list> cstrs
%type <Interfaces.constructor> cstr
%type <Interfaces.parameter list> params
%type <Interfaces.parameter list> preparam
%type <Interfaces.parameter list> indexparam
%type <(Interfaces.grammar * Interfaces.shift_exp) list -> Interfaces.parameter list> param
%type <(Interfaces.grammar * Interfaces.shift_exp) list> shift
%type <(Interfaces.grammar * Interfaces.shift_exp) list> shifts
%type <Interfaces.grammar * Interfaces.shift_exp> oneshift
%type <Interfaces.shift_exp> exp
%type <string> ident
%type <string list> idents

%%

gmod:
  MODULE ident DOT nodes END ident DOT
  { new gmodule_impl $2 [] $4 [] }
| MODULE ident DOT VERBATIM_HEADER nodes END ident DOT
  { new gmodule_impl $2 $4 $5 [] }
| MODULE ident DOT nodes VERBATIM_FOOTER END ident DOT
  { new gmodule_impl $2 [] $4 $5 }
| MODULE ident DOT VERBATIM_HEADER nodes VERBATIM_FOOTER END ident DOT
  { new gmodule_impl $2 $4 $5 $6 }

nodes:
  node
  { [ $1 ] }
| node nodes
  { $1 :: $2 }

node:
  INDUCTIVE grams DOT
  { new node_impl gmod $2 (current_pos ()) }

grams:
  ident COLON TYPE ASSIGN cstrs
  { new grammar_impl gmod $1 $5 (current_pos ()) :: [] }
| ident COLON TYPE ASSIGN cstrs WITH grams
  { new grammar_impl gmod $1 $5 (current_pos ()) :: $7 }

cstrs:
  cstr
  { [ $1 ] }
| cstr cstrs
  { $1 :: $2 }

cstr:
  BAR ident params
  { new constructor_impl gmod $2 $3 (current_pos ()) }
| BAR ident
  { new constructor_impl gmod $2 [] (current_pos ()) }

params:
  preparam
  { $1 }
| preparam params
  { $1 @ $2 }

preparam:
| shift param
  { $2 $1 }
| param
  { $1 [] }
| LPARSTAR INDEX STARRPAR indexparam
  { $4 }
| LPAR preparam RPAR
  { $2 }

indexparam:
  ident COLON ident
  { if $3 <> "nat" then Error.fail (current_pos ()) "index type must be nat"
    else [ new index_parameter_impl gmod $1
	     (new external_impl "nat" (current_pos ())) (current_pos ()) ] }

param:
  idents COLON ident
  { fun sh ->
    (List.map
       (fun id ->
	 new parameter_impl gmod sh id
	   (new external_impl $3 (current_pos ())) (current_pos ())) $1) }

shift:
  LPARSTAR BIND shifts IN STARRPAR
  { $3 }

shifts:
  oneshift
  { [ $1 ] }
| oneshift COMMA shifts
  { $1 :: $3 }

oneshift:
  ident
  { (new external_impl $1 (current_pos ()), SInt 1) }
| exp COLON ident
  { (new external_impl $3 (current_pos ()), $1) }

exp:
  INTEGER
    { SInt $1 }
| ident
    { SString $1 }
| exp PLUS exp
    { SPlus ($1, $3) }
| exp MINUS exp
    { SMinus ($1, $3) }
| exp MULT exp
    { SMult ($1, $3) }
| ident exp
    { SFun ($1, $2) }
| LPAR exp RPAR
    { $2 }

ident:
  ID
    { $1 }

idents:
  ident
    { [$1] }
| ident idents
    { $1 :: $2 }
